/*** 
	MACRO to implement Algorithm 4.1 of Westfall and Young (1993) 
	with respect to the MAX test of Zheng and Gastwirth (Statistics 
	in Medicine, Oct. 2006). 
	Developed by Dana Aeschliman <dana at statgen.org>
	for Marie-Pierre Dube's 
	Statistical Genetics Research Group at
	the Montreal Heart Institute 
	during Jan-June 2007
***/

%macro 	PREPARE4STEPDOWN(
dsn=, /* Input data set name */
aff_stat=, /* Variable for affected status */
locus=locus, /* Variable giving locus name. Can be "locus". Note 
	that locus names contain only numbers, letters and underscores.*/
A1=A1, /* Variable giving name of first allele. Can be "a1". */
A2=A2, /* Variable giving name of second allele. Can be "a2". */
num_resamps=10000,
/*  	Number of resampled data sets to create.  */
IDENT=id,
output_DSN=keep_max,
/*** output_DSN is the file that holds the single true and the many synthetic max_abs_zs ***/
use_lib=WORK,
/*** 	use_lib is used for several things.
	It's the library that holds the raw data. It's the library
	where the quality control (QC) files &use_lib..bad_ccs, 
	&use_lib..miss_or_diff_als, and &use_lib..snp_warnings 
	will be deposited. See this section in the MAXTEST macro 
	for more details about &use.lib ***/ 
keep_both=F,
/*** 	If this is 'T' then a copy of the data set both is saved in 
	use_lib ***/
skip_2_start2=F,
/*** 	If this is 'T' then program skips down to %start2. Note that, 
	when this is false, program assumes that it's OK to overwrite 
	&use_lib.output_file_names, i.e. program assumes that we are 
	just starting to build a set of data sets to be collated later
	when the step down algorithm is called.
***/
overwrite_output_files=F
);
%let dsn=%upcase(&dsn.); 
%let use_lib=%upcase(&use_lib.); 
%let keep_both=%upcase(&keep_both);
%let skip_2_start2=%upcase(&skip_2_start2.);
%let overwrite_output_files=%upcase(&overwrite_output_files.);
/*** 	The next 2 lines just tell SAS to read some MACROs, but nothing 
	will happen until those MACROs are called ***/
%include 'max_test_macro.sas';
%include 'DO_QC.sas';


%LET output_DSN=%UPCASE(&output_DSN.);
%LET use_lib=%UPCASE(&use_lib.);
/************************************************************************/
/*** 	If program will deposit output file name into 
	use_lib.output_file_names then:
		1) determine if use_lib.output_file_names exists 
		already and, if so,
		2) make sure that we don't overwrite an existing 
		file name 
***/
/*** 	Check if the proposed output file, 
	&output_DSN, already exists on &use_lib. As well, check that
	the file OUTPUT_FILE_NAMES, already exists on &use_lib. ***/

	%let _found_OUTPUT_FILE_NAMES_=F;
	%if %sysfunc(exist(&use_lib..OUTPUT_FILE_NAMES)) %then
		%let _found_OUTPUT_FILE_NAMES_=T;

%let _OUTPUT_DSN_ALREADY_EXISTS_=F;
%if %sysfunc(exist(&use_lib..&output_DSN)) %then
		%let _OUTPUT_DSN_ALREADY_EXISTS_=T;

	%let _OUTPUT_DSN_ALREADY_LISTED_=F;
/***
	If the file OUTPUT_FILE_NAMES already exists, determine 
	if the desired output file from the current iteration of this 
	MACRO is already listed in output_file_names.
***/
	%if &_found_OUTPUT_FILE_NAMES_=T %then %do;
		data tmp; set &use_lib..output_file_names ; 
		if upcase(trim(left(output_file)))="&use_lib..&output_DSN" 
			then do;
			call symput('_OUTPUT_DSN_ALREADY_LISTED_', 'T');
		end;
		run;
	%end;
/******************************************************************************/
/*** 
	In this section we make it difficult for the user to 
	overwrite data that has been produced previously.
***/
/*** 	If the user doesn't want to overwrite existing output... 
***/
%if &overwrite_output_files=F %then %do;
/*** 	If we are here, we know the following:
		1) the user doesn't want to overwrite existing output,

		2) if a file with the same name as that of the desired 
		output file from the current iteration of this MACRO 
		already exists on &use_lb or not,

		3) if the file OUTPUT_FILE_NAMES already exists on 
		&use_lib or not, and

		4) in the case where the file OUTPUT_FILE_NAMES already 
		exists on &use_lib, if a file with the same name as that 
		of the desired output file from the current iteration of 
		this MACRO is already listed in the file OUTPUT_FILE_NAMES.
***/
%if &_OUTPUT_DSN_ALREADY_EXISTS_=T AND &_found_OUTPUT_FILE_NAMES_=F 
	%then %do;
	%put ; 
	%put ********************************************************; 
	%put;
	%put A file named &output_DSN already exists on library &use_lib..;
	%put However, no file yet exists to hold the names of previously;
%put produced output files. Please either set overwrite_output_files to T;
%put or set output_DSN to something other than &output_DSN. and try again.;
	%put ; 
	%put ********************************************************; 
	%put;
	%goto done;
%end;
%if &_OUTPUT_DSN_ALREADY_EXISTS_=T AND &_found_OUTPUT_FILE_NAMES_=T
	%then %do;
	%put ; 
	%put ********************************************************; 
	%put;
	%put A file named &output_DSN already exists on library &use_lib..;
	%put As well, a file exists to hold the names of previously; 
	%if &_OUTPUT_DSN_ALREADY_LISTED_=T %then %do;
		%put produced output files. That file already lists &output_DSN.!! Do you;
		%put really wanna overwrite your previous output? If so, Please set;
		%put overwrite_output_files to T.;
		%put ; 
	%put ********************************************************; 
		%put;
	%end;
	%else %do;
		%put produced output files. But that file does not list &output_DSN.. Do you;
		%put  really wanna overwrite your previous output? If so, please set ;
		%put overwrite_output_files to T.;
		%put ; 
		%put ********************************************************; 
		%put;
		%end; 
		%goto done;
	%end;
	%if &_OUTPUT_DSN_ALREADY_EXISTS_=F AND &_found_OUTPUT_FILE_NAMES_=T and
		&_OUTPUT_DSN_ALREADY_LISTED_=T %then %do;
		%put ; 
		%put ********************************************************; 
		%put;
		%put A file named &output_DSN does not yet exist on library &use_lib. ;
		%put However, a file already exists to hold the names of previously;
		%put produced output files according to which &output_DSN. already exists!!;
		%put Please set overwrite_output_files to T and try again.;
		%put ; 
		%put ********************************************************; 
		%put;
		%goto done;
	%end;
%end; /*** %if &overwrite_output_files=F %then %do ***/

/******************************************************************************/
%let skip_2_start2=%upcase(&skip_2_start2.);
%if &skip_2_start2=T %then %goto start2;
/******************************************************************
********************************************************************/

%QC(dsn=&dsn, use_lib=&use_lib, FROM_MACRO=PREPARE4STEPDOWN, out_dsn=good_data);
%if &_abort_=1 %then %goto done;

/***  
The file "good_data" is in long format and has columns:
locus, a1, a2, CASE, ID, min_allele and maj_allele.
***/

%MAXTEST(dsn=good_data, aff_stat=case, locus=locus, A1=A1, A2=A2, 
IDENT=id, output_dsn=&output_DSN, use_lib=&use_lib, 
skip_checks=T, keep_details=T);
%start:
/****************************************************************************************/

/*** convert &dsn to short type so that we can resample it ***/

/****************************************************************************************/

proc sort data=good_data; by case id locus_no; run;

proc transpose data=good_data out=short;
var a1 a2; id locus_no ; by case id;
run;

/***  
The file "short" has columns:
CASE, ID, _NAME_, and then many columns for those SNPs which passed the QC checks.
Note that this set is sorted by case, id, then locus. 
***/
/*** See if any rows of short are missing observations ***/

data a1s(drop=_name_) a2s(drop=_name_ rename=(
	%do i = 1 %to &SNP_count; 
		_&i.=_X&i. 
	%end;
)); 
set short; 
if _name_="a1" then output a1s; 
if _name_="a2" then output a2s; run;;
run;

/*** Clean up datasets ***/;
proc datasets; delete tmp short good_data; run;

/**************************************************************************
Above, we put the transposed info into short, split off from short the
al1 allele and the al2 allele, and renamed the header for
allele 2.  

Next, we merge a1s and a2s to create the file both.

After that, we can start doing some iterations. 

**************************************************************************/;

proc sort data=a1s; by id; run;
proc sort data=a2s; by id; run;
data both; merge a1s(in=A) a2s(in=B); by id; if A and B; run;

/*** "both" has columns CASE, ID, then SNP names for the a1's (probably thousands)
followed by SNP names for the a2's which are the same as the SNP names for the a1's
but with the "X" prefixed. ***/
/*** Clean up datasets ***/;
proc datasets; delete a1s a2s; run;
/*** If needed, save a copy of 'both' ***/
%if &keep_both=T %then %do;
data &use_lib..both; set both; run;
%end;
/***************************************************************************/
%start2:
/*** If we just skipped to here from the beginning, we need to bring back 'both' ***/
%if &skip_2_start2=T %then %do;
/*** April 27, 2007: Should we do some QC on both? How do we know, in cases
where we are rerunning this, if it's the DS we want? ***/
data both; set &use_lib..both; run;
/*** Sort &use_lib..max_data because it will be immediately merged with 
give_to_max in MAXTEST. ***/
proc sort data=&use_lib..max_data; 
by locus_no; run;
/*** Make sure the MACRO var SNP_count is initialized ***/
data _NULL_; set &use_lib..params; call symput('SNP_count', SNP_count); run; 

%end; /*** END OF %if &skip_2_start2=T %then %do; ***/
/************************************************************************************
Start doing some iterations
************************************************************************************/
%do ii=1 %to &num_resamps;
/*** Draw a sample permutation of the cases on data set "both" ***/;
data tmp; set both(keep=case );
rand_rank=ranuni(-1);
*** Use the time of day as a seed;
run;
/*** sort them randomly ***/
proc sort data=tmp; by rand_rank;
run;
data both_CC_permd; merge both tmp(drop=rand_rank); run;
/*** Clean up datasets ***/;
proc datasets; delete tmp; run;

/*** Put data back into "long" format ***/;
proc transpose data=both_CC_permd out=transpd_permd; 

var %do i = 1 %to &SNP_count;
		_&i.
		_X&i.
	%end; 
;
by ID case;
run;

/*** Clean up data sets ***/
proc datasets; delete both_CC_permd; run;
%schmuck2:
/*** We now have the data back into long format but need to change the SNP names
back. ***/
data a1s(drop= _name_ rename=(col1=a1)) a2s(drop=_name_ rename=(col1=a2)); 
	set transpd_permd; 
if index(_name_, 'X')>0 then do;
	locus_no=input(substr(_name_, 3), best8.);
	output a2s; 
end;
else do;
	locus_no=input(substr(_name_, 2), best8.);
	output a1s;
end;
run;

/*** Clean up datasets ***/;
proc datasets; delete transpd_permd; run;

proc sort data=a1s; by locus_no ID; run;
proc sort data=a2s; by locus_no ID; run;
/*** Merge a1s and a2s, clean out rows w/ missing alleles
and add field "this_locus", an indicator which becomes "1"
for any locus where at least one subject has missing alleles. ***/
%if &ii=1 %then %do;
%global _hay_missing_;
%let _hay_missing_=0;
%end;

%if &ii=1 or &_hay_missing_=1 %then %do;
/*** If this is the first permutation or if there's missing data... ***/
data give_to_max(drop=hay_missing); 
merge a1s a2s end=last; 
by locus_no ID;
retain this_locus hay_missing 0;
if first.locus_no then this_locus=0;
/*** Need to clean out rows created where there are missing values ***/
if lengthn(a1)>0 then output give_to_max;
if lengthn(a1)=0 then do;
	this_locus=1;
	hay_missing=1;
end;
if last and hay_missing=1 then call symput('_hay_missing_', '1');
run;
%end;
%else %do; /*** This is a complete data set. ***/
data give_to_max; 
merge a1s a2s; 
by locus_no ID;
run;
%end;
/*** Clean up datasets ***/
proc datasets; delete a1s a2s ; run;
%start3:

%MAXTEST(dsn=give_to_max, aff_stat=case, locus=locus, A1=A1, A2=A2, 
IDENT=id, output_dsn=new_maxs, use_lib=&use_lib., skip_checks=T, keep_details=F);
/*** the default for &output_DSN is keep_max ***/; 
/*** if &ii=1 and &skip_2_start2=T then we are calling this program for at least 
the 2nd time in this analysis ***/
%if &ii=1 and &skip_2_start2=T %then %do;
	proc datasets library=work;
	modify new_maxs;
	change new_maxs=&output_DSN;
	run;
%end;
%else %do;
	proc append base=&output_DSN data=new_maxs;
	run;
%end;
/*** Clean up data sets ***/
	proc datasets ; delete give_to_max /*** new_maxs ***/; run;
%end; /*** end of %do ii=1 %to &num_resamps ***/
/*** Clean up data sets by deleting the DS 'both' ***/
proc datasets ; delete both ; run;
/************************************************************************/
/*** Convert to wide format ***/
proc sort data=&output_DSN; by locus_no; run;
proc transpose data=&output_DSN out=tmp(drop=_name_) prefix=MAX_; 
var max_abs_zs; by locus_no; 
run;
/*** We use the data step here because it'll work regardless of the library name. ***/
data &use_lib..&output_DSN; set tmp; run;
%if &use_lib^=WORK %then %do;
	proc datasets ; delete tmp &output_DSN; run;
%end;
%else %do;
	proc datasets ; delete tmp &output_DSN; run;
%end;
/************************************************************************/
/*** Deposit output file name into use_lib.output_file_names ***/
	%if &_found_OUTPUT_FILE_NAMES_=T and &skip_2_start2=T %then %do;
	/*** The MACRO variable _found_OUTPUT_FILE_NAMES_ is set at the start of
	this program. &_found_OUTPUT_FILE_NAMES_=T => output_file_names already exists, 
	so add to it ***/
		data tmp;
		length output_file $100;
		output_file="&use_lib..&output_DSN"; 
		num_resamps=&num_resamps;
		run;

		data &use_lib..output_file_names; 
		set &use_lib..output_file_names(where=(output_file^=upcase("&use_lib..&output_DSN"))) 
			tmp;
		run;
	%end;
	%else %do;
	/*** If output_file_names doesn't exist yet (&_found_OUTPUT_FILE_NAMES_^=T), 
	create it. Also, if  &skip_2_start2^=T, we must be starting a new round of generating
	null data, so it's OK to overwrite the DS &use_lib..output_file_names. ***/
		data &use_lib..output_file_names;
		length output_file $100;
		output_file="&use_lib..&output_DSN"; 
		num_resamps=&num_resamps;
		run;
	%end; /*** ***/
%done:
%MEND; /* END OF MACRO STEPDOWN4SNPS */














